///////////////////////////////////////////////////////////////////////////////
// nonlinear.C: 
//
// Copyright (C) 2001 <--> $Date: 2018/11/26 01:55:50 $, Hugh Blackburn.
//
// This version of basic semtex DNS includes transport of a passive
// scalar 'c', carried as the NCOMth Field; pressure is
// NCOM+1. Cartesian and cylindrical coordinates.
///////////////////////////////////////////////////////////////////////////////

static char RCS[] = "$Id: nonlinear.cpp,v 1.7 2018/11/26 01:55:50 hmb Exp $";

#include <scat.h>


void nonlinear (Domain*         D ,
		BCmgr*          B ,
		AuxField**      Us,
		AuxField**      Uf,
		vector<real_t>& ff,
		vector<real_t>& g )
// ---------------------------------------------------------------------------
// Mainly: compute advection terms in Navier--Stokes equations: N(u).
// Since they involve product terms they are computed pseudo-
// spectrally (in physical space): anything that involves products of
// field variables should be done here.  Terms of N are computed in
// physical space, then forward Fourier transformed.
//
// Velocity field data areas of D and first level of Us are swapped, then
// the next stage of nonlinear forcing terms N(u) are computed from
// velocity fields and left in the first level of Uf.
//
// Nonlinear terms N(u) are computed in skew-symmetric form (Zang 1991)
//                 ~ ~
//                   N = -0.5 ( u . grad u + div uu )
//                   ~          ~        ~       ~~
//
// i.e., in Cartesian component form
//
//              N  = -0.5 ( u  d(u ) / dx  + d(u u ) / dx ).
//               i           j    i      j      i j      j
//
// Though in fact we use a cheaper approximation here -- alternating
// skew symmetric -- alternately u.grad(u) and div(uu) are made on
// successive timesteps.  Bob Kerr seems to have invented this idea
// way back when.
//
// Note that in cylindrical coordinates, the axial and radial
// nonlinear terms are constructed pre-multiplied by radius (y).  See
// Blackburn & Sherwin JCP 2004 for am explanation.
//
// The temperature transport term u . grad c is also built here, in
// N[NCOM].
//
// Finally, this routine is used to calculate Boussinesq buoyancy
// terms both of gravitational and (optionally) centrifugal types (see
// Lopez, Marques & Avila JFM 737, 2012, for a good introduction).
//
// 1. Gravitational buoyancy
//
// The (gravitational) Boussinesq buoyancy term (when on the RHS of)
// the momentum equation is
//
//                      - BETA_T * (T - T_REF) g.
//                                             ~
// We add an explicit estimate of this to the nonlinear terms in Uf.
// The first level of Us has the last values of the data Fields, D is free.
//
// Cylindrical coordinates: for simplicity, we only presently allow
// axial gravitational buoyancy.  Off-axis gravity could be dealt with
// without too much difficulty.
//
// 2. Centrifugal buoyancy.
//
// The "centrifugal" Boussinesq buoyancy term (again on the RHS of)
// the momentum equation is
//
//                     + BETA_T * (T - T_REF) * N(u).
//                                              ~ ~
// 
// ---------------------------------------------------------------------------
{
  const int_t NDIM = Geometry::nDim();	// -- Number of space dimensions.
  const int_t NCOM = D -> nField() - 2;	// -- Number of velocity components.
  const int_t nP   = Geometry::planeSize();
  const int_t nZ   = Geometry::nZ();
  const int_t nZP  = Geometry::nZProc();
  const int_t nTot = Geometry::nTotProc();

  vector<AuxField*> U (NCOM + 1), N (NCOM + 1); 
  Field*            master = D -> u[0];
  int_t             i, j;
  static int               toggle = 1; // -- Switch u.grad(u) or div(uu).
  static real_t*           work;
  static vector<AuxField*> Uphys (NCOM + 1);
  static AuxField*         tmp;

  if (!work) { 			// -- First time through.
    work = new real_t [static_cast<size_t>((NCOM + 2) * nTot)];
    for (i = 0; i <= NCOM; i++)
      Uphys[i] = new AuxField (work + i * nTot, nZP, D -> elmt);
    tmp = new AuxField (work + (NCOM + 1) * nTot, nZP, D -> elmt);
  }

  for (i = 0; i <= NCOM; i++) {
    AuxField::swapData (D -> u[i], Us[i]);
    U[i] = Us[i];
    N[i] = Uf[i];
    *N[i] = 0.0;
    (*Uphys[i] = *U[i]) . transform (INVERSE);
  }

  B -> maintainPhysical(master, Uphys, NCOM);

  if (Geometry::cylindrical()) {

    if (toggle) { // -- Convective component u.grad(u).

      for (i = 0; i <= NCOM; i++) {

	// -- Terms involving azimuthal derivatives and frame components.

	if (NCOM > 2) {
	  if (i == 1) N[1] -> times      (*Uphys[2], *Uphys[2]);
	  if (i == 2) N[2] -> timesMinus (*Uphys[2], *Uphys[1]);

	  if (nZ > 2) {
	    (*tmp = *U[i]) . gradient (2) . transform (INVERSE);
	    N[i] -> timesMinus (*Uphys[2], *tmp);
	  }
	}

	if (i >= 2) N[i] -> divY ();

	// -- 2D convective derivatives.

	for (j = 0; j < 2; j++) {
	  (*tmp = *Uphys[i]) . gradient (j);
	  if (i < 2) tmp -> mulY ();
	  N[i] -> timesMinus (*Uphys[j], *tmp);
	}
      }
      
    } else { // -- Conservative component div(uu).

      for (i = 0; i <= NCOM; i++) {

	// -- Terms involving azimuthal derivatives and frame components.

	if (i == 0) N[0] -> timesMinus (*Uphys[0], *Uphys[1]);
	if (i == 1) N[1] -> timesMinus (*Uphys[1], *Uphys[1]);

	if (NCOM > 2) {

	  if (i == 1) N[1] -> timesPlus (*Uphys[2], *Uphys[2]);

	  if (i == 2) {
	    tmp -> times (*Uphys[2], *Uphys[1]);
	    N[2] -> axpy (-2., *tmp);
	  }

	  if (i == 3) N[3] -> timesMinus (*Uphys[1], *Uphys[3]);

	  if (nZ > 2) {
	    tmp -> times (*Uphys[i], *Uphys[2]);
	    (*tmp) . transform (FORWARD). gradient (2). transform (INVERSE);
	    *N[i] -= *tmp;
	  }
	}

	if (i >= 2) N[i] -> divY ();

	// -- 2D conservative derivatives.
     
	for (j = 0; j < 2; j++) {
	  (*tmp). times (*Uphys[j], *Uphys[i]) . gradient (j);
	  if (i < 2) tmp -> mulY ();
	  *N[i] -= *tmp;
	}
      }
    }

    // -- Deal with body force terms (same for either formulation of NL):
    
    if (Femlib::ivalue ("CENT_BUOY")) { // -- Centrifugal buoyancy.
      *tmp  = *Uphys[NCOM];
      *tmp -=  Femlib::value ("T_REF");
      *tmp *= -Femlib::value ("BETA_T");
      *tmp += 1.0;

      for (i = 0; i < NCOM; i++) *N[i] *= *tmp;
    }

    if (fabs (g[0]) > EPSDP) {       // -- Gravitational buoyancy (axial only).
      *tmp = *Uphys[NCOM];
      *tmp -=  Femlib::value ("T_REF");
      *tmp *= -Femlib::value ("BETA_T") * g[0];
      tmp -> mulY ();
      *N[0] += *tmp;
    }

    // -- Constant body force.

    for (i = 0; i < NCOM; i++) {
      if (fabs (ff[i]) > EPSDP) {
	*tmp = ff[i];
	if (i < 2) tmp -> mulY ();
	*N[i] += *tmp;
      }
    }
    
  } else {			// -- Cartesian coordinates.

    if (toggle) { // -- Convective component u.grad(u).

      for (i = 0; i <= NCOM; i++) {
	for (j = 0; j < NDIM; j++) {
      
	  // -- Perform n_i -= u_j d(u_i) / dx_j.

	  if (j == 2) (*tmp = *U[i]) . gradient (j) . transform (INVERSE);
	  else    (*tmp = *Uphys[i]) . gradient (j);
	  N[i] -> timesMinus (*Uphys[j], *tmp);
	}
      }

    } else { // -- Conservative component div(uu).

      for (i = 0; i <= NCOM; i++) {
	for (j = 0; j < NDIM; j++) {

	  // -- Perform n_i -= d(u_i u_j) / dx_j.

	  tmp -> times (*Uphys[i], *Uphys[j]);
	  if (j == 2) tmp -> transform (FORWARD);
	  tmp -> gradient (j);
	  if (j == 2) tmp -> transform (INVERSE);
	  *N[i] -= *tmp;
	}
      }
    }

    // -- Deal with body force terms (same for either formulation of NL):
    
    if (Femlib::ivalue ("CENT_BUOY")) { // -- Centrifugal buoyancy.
      *tmp  = *Uphys[NCOM];
      *tmp -=  Femlib::value ("T_REF");
      *tmp *= -Femlib::value ("BETA_T");
      *tmp += 1.0;

      for (i = 0; i < NCOM; i++) *N[i] *= *tmp;
    }

    for (i = 0; i < NCOM; i++) {        // -- Gravitational buoyancy.
      if (fabs (g[i]) > EPSDP) {
	*tmp = *Uphys[NCOM];
	*tmp -= Femlib::value ("T_REF");
	*tmp *= Femlib::value ("BETA_T") * g[i];
	*N[i] -= *tmp;
      }
    }

    for (i = 0; i < NCOM; i++)	        // -- Constant body force.
      if (fabs (ff[i]) > EPSDP) *N[i] += ff[i];

  }

  // -- Back to Fourier space.
    
  for (i = 0; i <= NCOM; i++) {
    master -> smooth (N[i]);
    N[i] -> transform (FORWARD);
  }
 
  toggle = 1 - toggle;
}

#if 0

// -- No longer required since it's all in nonlinear.  For reference.

static void buoyancy (Domain*         D ,
		      AuxField**      Us,
		      AuxField**      Uf,
		      vector<real_t>& g )
// ---------------------------------------------------------------------------

// ---------------------------------------------------------------------------
{
  int_t     i;
  AuxField* T    = Us [NCOM];
  AuxField* work = D -> u[NCOM];

  if (Geometry::cylindrical()) {
    if (fabs (g[0]) > EPSDP) {
      *work      = *T;
      ROOTONLY work -> addToPlane (0, -Femlib::value ("T_REF"));
      *work     *=  Femlib::value ("BETA_T");
      *work     *=  g[0];
       work     -> mulY();
      *Uf[0] -= *work;
    }
  } else
    for (i = 0; i < NCOM; i++)
      if (fabs (g[i]) > EPSDP) {
	*work      = *T;
	ROOTONLY work -> addToPlane (0, -Femlib::value ("T_REF"));
	*work     *=  Femlib::value ("BETA_T");
	*work     *=  g[i];
	*Uf[i] -= *work;
      }
}

#endif
